/*
backlash.c - Backlash compensation module
  Part of DiyouPCBFirm

Copyright (c) 2013 V.Ayllon.  All rights reserved.
 
Inspired in RepRap NoLash Software BackLash Compensation Tool
See http://forums.reprap.org/read.php?154,178612,182380#msg-182380

DiyouPCBFirm is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the GNU General Public License, Version 2 or later, for full details.

Everyone is granted permission to copy, modify and redistribute
DiyouPCBFirm, but only under the conditions described in the GNU General Public
License.  A copy of this license is supposed to have been
given to you along with DiyouPCB so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

#include "backlash.h"
#include "DiyouPCBFirm.h"
#include "planner.h"
#include "stepper.h"

float currentx, currenty;


/** PlanBufferLine()
 ** Add a new linear movement to the buffer. steps_x, _y and _z is the absolute position in 
 ** mm. Microseconds specify how many microseconds the move should take to perform. To aid acceleration
 ** calculation the caller must also provide the physical length of the line in millimeters.
 ** laserState defines if the laser is on or off.
 ** introduce backlash compensation if needed.
 */
void planBufferLine(const float &x, const float &y, float feed_rate, bool laserState, int laserPower, int laserEntry, int laserExit){


	currentx = sys.current_position[X_AXIS];
	currenty = sys.current_position[Y_AXIS];


	// Check if we have to insert an extra movement in X
	// first we check if the axis has some change. Then if it changed direction and lash is not cero.
	//

	if (!isSameDirection(X_AXIS, x) && sys.backlash[X_AXIS] > 0.0f) {
		sys.current_dir[X_AXIS] = -sys.current_dir[X_AXIS]; // reverse the direction of movement
		currentx += (sys.backlash[X_AXIS]/2.0f) * sys.current_dir[X_AXIS]; // calculate countermeasure position
	}

	// Check if we have to insert an extra movement in Y
	// first we check if the axis has some change. Then if it changed direction and lash is not cero.
	//

	if (!isSameDirection(Y_AXIS, y) && sys.backlash[Y_AXIS] > 0.0f) {
		sys.current_dir[Y_AXIS] = -sys.current_dir[Y_AXIS]; // reverse the direction of movement
		currenty += (sys.backlash[Y_AXIS]/2.0f) * sys.current_dir[Y_AXIS]; // calculate countermeasure position
	}
	// If any axis changed, then queue an extra buffer line
	// LaserPower =  laserEntry to avoid too much power
	//
	if (currentx != sys.current_position[X_AXIS] || currenty != sys.current_position[Y_AXIS]){
		queueBufferLine(currentx, currenty, feed_rate, laserState, laserEntry, laserEntry, laserExit);
		sys.current_position[X_AXIS] = currentx;
		sys.current_position[Y_AXIS] = currenty;
	}
	//
	// Now insert the normal buffer line also with compensation at the target
	// and update the position.
	//
	currentx = x + (sys.backlash[X_AXIS]/2.0f) * sys.current_dir[X_AXIS]; // save recalculated position
	currenty = y + (sys.backlash[Y_AXIS]/2.0f) * sys.current_dir[Y_AXIS]; // save recalculated position
	queueBufferLine(currentx, currenty, feed_rate, laserState, laserPower, laserEntry, laserExit);
	// Update position is needed for circles because are made of a lot small segments
	sys.current_position[X_AXIS] = currentx;
	sys.current_position[Y_AXIS] = currenty;
}


/**
** isSameDirection() - Compares the two positions and determines the direction of movement.
**/
bool isSameDirection(int axis, float target) {
    int newDir = 0;

    if (sys.current_position[axis] < target) {
		newDir=1;
    } else if (sys.current_position[axis] > target) {
      	newDir=-1;
    } else {
      return true;
    }
    return (newDir == sys.current_dir[axis]);
}
