/*
  motion_control.c - high level interface for issuing motion commands
 Part of Grbl
 
 Copyright (c) 2009-2011 Simen Svale Skogsrud
 Copyright (c) 2011 Sungeun K. Jeon
 
 Grbl is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Grbl is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Grbl.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <math.h>
#include "DiyouPCBFirm.h"
//#include "stepper.h"
//#include "planner.h"
#include "backlash.h" 

// The arc is approximated by generating a huge number of tiny, linear segments. The length of each 
// segment is configured in settings.mm_per_arc_segment.  
bool mc_arc(float *position, float *target, float *offset, float feed_rate, float radius, bool isclockwise,bool laserState, int laserPower)
{   

  //   int acceleration_manager_was_enabled = plan_is_acceleration_manager_enabled();
  //   plan_set_acceleration_manager_enabled(false); // disable acceleration management for the duration of the arc
  float center_axis0 = position[X_AXIS] + offset[X_AXIS];
  float center_axis1 = position[Y_AXIS] + offset[Y_AXIS];
  float r_axis0 = -offset[X_AXIS];  // Radius vector from center to current location
  float r_axis1 = -offset[Y_AXIS];
  float rt_axis0 = target[X_AXIS] - center_axis0;
  float rt_axis1 = target[Y_AXIS] - center_axis1;

  int segments;
  float angular_travel;
  int laserEntry, laserExit;

  // CCW angle between position and target from circle center. Only one atan2() trig computation required.
  angular_travel = atan2(r_axis0*rt_axis1-r_axis1*rt_axis0, r_axis0*rt_axis0+r_axis1*rt_axis1);
  
   if (isclockwise) { // Correct atan2 output per direction
    if (angular_travel >= -0.001) { 
      angular_travel -= 2*M_PI;
    }
  } 
  else {
    if (angular_travel <= 0.0001) { 
      angular_travel += 2*M_PI;
    }
  }
  //float millimeters_of_travel = angular_travel*radius;  Bugg--

  float millimeters_of_travel = hypot(angular_travel*radius,0.0); 
 // float millimeters_of_travel = abs(angular_travel)*radius; 
 
  if (millimeters_of_travel < 0.001) { 
    return false;
  }

  if (radius <= SMALL_CIRCLE_RADIUS){
    segments = floor(millimeters_of_travel/MM_PER_ARC_SEGMENT_SMALL);
 	feed_rate = SMALL_CIRCLE_FEEDRATE;
	laserPower = SMALL_CIRCLE_POWER;
	laserEntry = SMALL_CIRCLE_ENTRY_POWER;
	laserExit = SMALL_CIRCLE_EXIT_POWER;
  }
  else {
    segments = floor(millimeters_of_travel/MM_PER_ARC_SEGMENT_LARGE);
	laserEntry = DEFAULT_ENTRY_POWER;
	laserExit = DEFAULT_EXIT_POWER;
  }
 
  if(segments == 0) segments = 1;

  float theta_per_segment = angular_travel/segments;


  /* Vector rotation by transformation matrix: r is the original vector, r_T is the rotated vector,
   and phi is the angle of rotation. Based on the solution approach by Jens Geisler.
   r_T = [cos(phi) -sin(phi);
   sin(phi)  cos(phi] * r ;
   
   For arc generation, the center of the circle is the axis of rotation and the radius vector is 
   defined from the circle center to the initial position. Each line segment is formed by successive
   vector rotations. This requires only two cos() and sin() computations to form the rotation
   matrix for the duration of the entire arc. Error may accumulate from numerical round-off, since
   all double numbers are single precision on the Arduino. (True double precision will not have
   round off issues for CNC applications.) Single precision error can accumulate to be greater than
   tool precision in some cases. Therefore, arc path correction is implemented. 
   
   Small angle approximation may be used to reduce computation overhead further. This approximation
   holds for everything, but very small circles and large mm_per_arc_segment values. In other words,
   theta_per_segment would need to be greater than 0.1 rad and N_ARC_CORRECTION would need to be large
   to cause an appreciable drift error. N_ARC_CORRECTION~=25 is more than small enough to correct for 
   numerical drift error. N_ARC_CORRECTION may be on the order a hundred(s) before error becomes an
   issue for CNC machines with the single precision Arduino calculations.
   
   This approximation also allows mc_arc to immediately insert a line segment into the planner 
   without the initial overhead of computing cos() or sin(). By the time the arc needs to be applied
   a correction, the planner should have caught up to the lag caused by the initial mc_arc overhead. 
   This is important when there are successive arc motions. 
  */
  // Vector rotation matrix values
  float cos_T = 1-0.5*theta_per_segment*theta_per_segment; // Small angle approximation
  float sin_T = theta_per_segment;

  float arc_target[2];
  float sin_Ti;
  float cos_Ti;
  float r_axisi;
  int i;
  int8_t count = 0;
  int lp,le,lx;
  lp = laserPower;
  le = laserEntry;
  lx = laserExit;

  for (i = 1; i<segments; i++) { // Increment (segments-1)

	if (i == 1){
		le = laserEntry;
		lp = lx = laserPower;
	} else if (i == (segments -1)){
		le = lp = laserPower;
		lx = laserExit;
	} else {
		lp = le = lx = laserPower;
	}

    if (count < N_ARC_CORRECTION) {
      // Apply vector rotation matrix 
      r_axisi = r_axis0*sin_T + r_axis1*cos_T;
      r_axis0 = r_axis0*cos_T - r_axis1*sin_T;
      r_axis1 = r_axisi;
      count++;
    } 
    else {
      // Arc correction to radius vector. Computed only every N_ARC_CORRECTION increments.
      // Compute exact location by applying transformation matrix from initial radius vector(=-offset).
      cos_Ti = cos(i*theta_per_segment);
      sin_Ti = sin(i*theta_per_segment);
      r_axis0 = -offset[X_AXIS]*cos_Ti + offset[Y_AXIS]*sin_Ti;
      r_axis1 = -offset[X_AXIS]*sin_Ti - offset[Y_AXIS]*cos_Ti;
      count = 0;
    }

    // Update arc_target location
    arc_target[X_AXIS] = center_axis0 + r_axis0;
    arc_target[Y_AXIS] = center_axis1 + r_axis1;

    clamToSoftwareEnstops(arc_target);

    planBufferLine(arc_target[X_AXIS], arc_target[Y_AXIS], feed_rate,laserState,lp,le,lx);
  }

  // Ensure last segment arrives at target location.
  planBufferLine(target[X_AXIS], target[Y_AXIS], feed_rate,laserState,lp,lp,lx);

  return true;
}
